from __future__ import annotations

import array
import enum
import itertools
from dataclasses import dataclass, field
from enum import Enum
from typing import ByteString, Optional


class SeatStatus(Enum):
    """
    Internal representation of the "status" of a seat, i.e. whether it's available to
    assign to guests.
    All values of this enum are integers to be able to put them in an `array.array`.
    """

    # There's no seat here
    NoSeat = ord(" ")

    # Seat can be assigned to a group
    Available = ord("o")

    # Seat has been taken
    Taken = ord("i")

    # Seat can't be used because it's within block range of one or more other groups.
    Blocked = ord("x")


@dataclass
class Cinema:
    width: int
    height: int
    seats: array.array[int] = field(init=False)

    def __post_init__(self) -> None:
        if self.width <= 0:
            raise ValueError(f"Width must be a positive integer, is {self.width}")
        if self.height <= 0:
            raise ValueError(f"Height must be a positive integer, is {self.height}")

        self.seats = array.array(
            "b", itertools.repeat(SeatStatus.NoSeat.value, self.width * self.height)
        )

    def _xy_to_index(self, x: int, y: int) -> int:
        if not 0 <= x < self.width:
            raise KeyError(f"`x` out of range: got {x}, width {self.width}")
        if not 0 <= y < self.height:
            raise KeyError(f"`y` out of range: got {y}, width {self.height}")

        return x + self.width * y

    def get_seat(self, x: int, y: int) -> SeatStatus:
        value = self.seats[self._xy_to_index(x, y)]
        return SeatStatus(value)

    def set_seat(self, x: int, y: int, new_status: SeatStatus) -> None:
        self.seats[self._xy_to_index(x, y)] = new_status.value

    def can_seat_group(self, x: int, y: int, group_size: int) -> bool:
        """
        Returns wether you can seat a group of this size to the
        right starting at the coordinates (x, y).
        """
        for check_x in range(x, x + group_size):
            if self.get_seat(check_x, y) != SeatStatus.Available:
                return False
        return True

    def seat_group(self, x: int, y: int, group_size: int) -> None:
        """
        Seats a group in the cinema to the right from the given
        coordinates. Throws an AssertionError if the group can't be
        seated here.
        """
        if not self.can_seat_group(x, y, group_size):
            raise AssertionError("Can't seat a group here")

        # Block seats (including the seats that we will use, for
        # convenience)
        for block_y in range(y - 1, y + 2):
            start_x = x
            end_x = x + group_size

            if block_y == y:
                start_x -= 2
                end_x += 2

            for block_x in range(start_x, end_x):
                if self.get_seat(block_x, block_y) == SeatStatus.Available:
                    self.set_seat(block_x, block_y, SeatStatus.Blocked)

        # Give the people their seats!
        for seat_x in range(x, x + group_size):
            self.set_seat(seat_x, y, SeatStatus.Taken)

    def get_amount_people_seated(self) -> int:
        """
        Returns the amount of people seated in the cinema
        """
        amount = 0
        for x in range(self.width):
            for y in range(self.height):
                if self.get_seat(x, y) == SeatStatus.Taken:
                    amount += 1

        return amount
